@@ -30,11 +30,12 @@
 #include "fm-tree-model.h"
 
 #include <eel/eel-glib-extensions.h>
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 #include <libnautilus-private/nautilus-directory.h>
 #include <libnautilus-private/nautilus-file-attributes.h>
 #include <libnautilus-private/nautilus-file.h>
 #include <libnautilus-private/nautilus-icon-factory.h>
+#include <gtk/gtkenums.h>
 #include <string.h>
 
 enum {
@@ -85,7 +86,7 @@ struct TreeNode {
 
 	/* misc. flags */
 	guint done_loading : 1;
-	guint inserting_first_child : 1;
+	guint force_has_dummy : 1;
 	guint inserted : 1;
 };
 
@@ -108,8 +109,6 @@ struct FMTreeModelRoot {
 	GHashTable *file_to_node_map;
 	
 	TreeNode *root_node;
-
-	gulong changed_handler_id;
 };
 
 typedef struct {
@@ -246,12 +245,12 @@ tree_node_get_pixbuf_from_factory (TreeNode *node,
 				   const char *modifier)
 {
 	if (node->parent == NULL) {
-		return nautilus_icon_factory_get_pixbuf_from_name
+		return nautilus_icon_factory_get_pixbuf_from_name_with_stock_size
 			(node->icon_name, NULL,
-			 NAUTILUS_ICON_SIZE_FOR_MENUS, NULL);
+			 GTK_ICON_SIZE_MENU, NULL);
 	}
-	return nautilus_icon_factory_get_pixbuf_for_file_force_size
-		(node->file, modifier, NAUTILUS_ICON_SIZE_FOR_MENUS);
+	return nautilus_icon_factory_get_pixbuf_for_file_with_stock_size
+		(node->file, modifier, GTK_ICON_SIZE_MENU);
 }
 
 static gboolean
@@ -304,9 +303,9 @@ tree_node_get_emblem_pixbuf_from_factory (TreeNode *node)
 	eel_string_list_free (emblems_to_ignore);
 
 	if (emblem_icons != NULL) {
-		pixbuf = nautilus_icon_factory_get_pixbuf_for_icon
+		pixbuf = nautilus_icon_factory_get_pixbuf_for_icon_with_stock_size
 			(emblem_icons->data, NULL,
-			 NAUTILUS_ICON_SIZE_FOR_MENUS,
+			 GTK_ICON_SIZE_MENU,
 			 NULL, NULL, FALSE, NULL);
 	} else {
 		pixbuf = NULL;
@@ -396,10 +395,12 @@ tree_node_get_display_name (TreeNode *node)
 static gboolean
 tree_node_has_dummy_child (TreeNode *node)
 {
-	return node->directory != NULL
+	return (node->directory != NULL
 		&& (!node->done_loading
 		    || node->first_child == NULL
-		    || node->inserting_first_child);
+		    || node->force_has_dummy)) ||
+		/* Roots always have dummy nodes if directory isn't loaded yet */
+		(node->directory == NULL && node->parent == NULL);
 }
 
 static int
@@ -612,15 +613,14 @@ report_dummy_row_deleted (FMTreeModel *model, TreeNode *parent)
 	GtkTreeIter iter;
 	GtkTreePath *path;
 
-	abandon_dummy_row_ref_count (model, parent);
-	if (!parent->inserted) {
-		return;
+	if (parent->inserted) {
+		make_iter_for_node (parent, &iter, model->details->stamp);
+		path = gtk_tree_model_get_path (GTK_TREE_MODEL (model), &iter);
+		gtk_tree_path_append_index (path, 0);
+		gtk_tree_model_row_deleted (GTK_TREE_MODEL (model), path);
+		gtk_tree_path_free (path);
 	}
-	make_iter_for_node (parent, &iter, model->details->stamp);
-	path = gtk_tree_model_get_path (GTK_TREE_MODEL (model), &iter);
-	gtk_tree_path_append_index (path, 0);
-	gtk_tree_model_row_deleted (GTK_TREE_MODEL (model), path);
-	gtk_tree_path_free (path);
+	abandon_dummy_row_ref_count (model, parent);
 }
 
 static void
@@ -632,7 +632,8 @@ report_node_inserted (FMTreeModel *model, TreeNode *node)
 	report_row_inserted (model, &iter);
 	node->inserted = TRUE;
 
-	if (node->directory != NULL) {
+	if (node->directory != NULL ||
+	    node->parent == NULL) {
 		report_row_has_child_toggled (model, &iter);
        }
        if (tree_node_has_dummy_child (node)) {
@@ -731,10 +732,11 @@ destroy_node (FMTreeModel *model, TreeNode *node)
 
 	path = get_node_path (model, node);
 
-	destroy_node_without_reporting (model, node);
-
+	/* Report row_deleted before actually deleting */
 	gtk_tree_model_row_deleted (GTK_TREE_MODEL (model), path);
 	gtk_tree_path_free (path);
+	
+	destroy_node_without_reporting (model, node);
 
 	if (tree_node_has_dummy_child (parent)) {
 		if (!parent_had_dummy_child) {
@@ -784,9 +786,11 @@ update_node_without_reporting (FMTreeModel *model, TreeNode *node)
 
 	changed = FALSE;
 	
-	if (node->directory == NULL && nautilus_file_is_directory (node->file)) {
+	if (node->directory == NULL &&
+	    (nautilus_file_is_directory (node->file) || node->parent == NULL)) {
 		node->directory = nautilus_directory_get_for_file (node->file);
-	} else if (node->directory != NULL && !nautilus_file_is_directory (node->file)) {
+	} else if (node->directory != NULL &&
+		   !(nautilus_file_is_directory (node->file) || node->parent == NULL)) {
 		stop_monitoring_directory (model, node);
 		destroy_children (model, node);
 		nautilus_directory_unref (node->directory);
@@ -808,7 +812,8 @@ insert_node (FMTreeModel *model, TreeNode *parent, TreeNode *node)
 
 	parent_empty = parent->first_child == NULL;
 	if (parent_empty) {
-		parent->inserting_first_child = TRUE;
+		/* Make sure the dummy lives as we insert the new row */
+		parent->force_has_dummy = TRUE;
 	}
 
 	tree_node_parent (node, parent);
@@ -817,9 +822,13 @@ insert_node (FMTreeModel *model, TreeNode *parent, TreeNode *node)
 	report_node_inserted (model, node);
 
 	if (parent_empty) {
-		parent->inserting_first_child = FALSE;
+		parent->force_has_dummy = FALSE;
 		if (!tree_node_has_dummy_child (parent)) {
+			/* Temporarily set this back so that row_deleted is
+			 * sent before actually removing the dummy child */
+			parent->force_has_dummy = TRUE;
 			report_dummy_row_deleted (model, parent);
+			parent->force_has_dummy = FALSE;
 		}
 	}
 }
@@ -838,11 +847,12 @@ reparent_node (FMTreeModel *model, TreeNode *node)
 
 	path = get_node_path (model, node);
 
-	abandon_node_ref_count (model, node);
-	tree_node_unparent (model, node);
-
+	/* Report row_deleted before actually deleting */
 	gtk_tree_model_row_deleted (GTK_TREE_MODEL (model), path);
 	gtk_tree_path_free (path);
+	
+	abandon_node_ref_count (model, node);
+	tree_node_unparent (model, node);
 
 	insert_node (model, new_parent, node);
 }
@@ -906,7 +916,11 @@ update_node (FMTreeModel *model, TreeNode *node)
 		if (has_dummy_child) {
 			report_dummy_row_inserted (model, node);
 		} else {
+			/* Temporarily set this back so that row_deleted is
+			 * sent before actually removing the dummy child */
+			node->force_has_dummy = TRUE;
 			report_dummy_row_deleted (model, node);
+			node->force_has_dummy = FALSE;
 		}
 	}
 	if (had_directory != has_directory) {
@@ -978,7 +992,11 @@ set_done_loading (FMTreeModel *model, TreeNode *node, gboolean done_loading)
 		}
 	} else {
 		if (had_dummy) {
+			/* Temporarily set this back so that row_deleted is
+			 * sent before actually removing the dummy child */
+			node->force_has_dummy = TRUE;
 			report_dummy_row_deleted (model, node);
+			node->force_has_dummy = FALSE;
 		} else {
 			g_assert_not_reached ();
 		}
@@ -995,7 +1013,14 @@ done_loading_callback (NautilusDirectory *directory,
 
 	file = nautilus_directory_get_corresponding_file (directory);
 	node = get_node_from_file (root, file);
-	g_assert (node != NULL);
+	if (node == NULL) {
+		/* This can happen for non-existing files as tree roots,
+		 * since the directory <-> file object relation gets
+		 * broken due to nautilus_directory_remove_file()
+		 * getting called when i/o fails.
+		 */
+		return;
+	}
 	set_done_loading (root->model, node, TRUE);
 	nautilus_file_unref (file);
 	
@@ -1185,10 +1210,12 @@ static void
 fm_tree_model_get_value (GtkTreeModel *model, GtkTreeIter *iter, int column, GValue *value)
 {
 	TreeNode *node, *parent;
+	FMTreeModel *fm_model;
 
 	g_return_if_fail (FM_IS_TREE_MODEL (model));
 	g_return_if_fail (iter_is_valid (FM_TREE_MODEL (model), iter));
 	
+	fm_model = FM_TREE_MODEL (model);
 	node = iter->user_data;
 
 	switch (column) {
@@ -1303,7 +1330,7 @@ fm_tree_model_iter_has_child (GtkTreeModel *model, GtkTreeIter *iter)
 
 	node = iter->user_data;
 
-	has_child = node != NULL && node->directory != NULL;
+	has_child = node != NULL && (node->directory != NULL || node->parent == NULL);
 
 #if 0
 	g_warning ("Node '%s' %s",
@@ -1517,20 +1544,11 @@ fm_tree_model_unref_node (GtkTreeModel *model, GtkTreeIter *iter)
 	}
 }
 
-static void
-root_node_file_changed_callback (NautilusFile *file, FMTreeModelRoot *root)
-{
-	if (root->root_node != NULL) {
-		update_node (root->model, root->root_node);
-	}
-}
-
 void
 fm_tree_model_add_root_uri (FMTreeModel *model, const char *root_uri, const char *display_name, const char *icon_name, GnomeVFSVolume *volume)
 {
 	NautilusFile *file;
 	TreeNode *node, *cnode;
-	NautilusFileAttributes attributes;
 	FMTreeModelRoot *newroot;
 	
 	file = nautilus_file_get (root_uri);
@@ -1554,13 +1572,6 @@ fm_tree_model_add_root_uri (FMTreeModel *model, const char *root_uri, const char
 		node->prev = cnode;
 	}
 
-	newroot->changed_handler_id = g_signal_connect (node->file, "changed",
-							G_CALLBACK (root_node_file_changed_callback),
-							node->root);
-
-	attributes = get_tree_monitor_attributes ();
-	nautilus_file_monitor_add (file, model, attributes);
-
 	nautilus_file_unref (file);
 
 	update_node_without_reporting (model, node);
@@ -1612,6 +1623,10 @@ fm_tree_model_remove_root_uri (FMTreeModel *model, const char *uri)
 		nautilus_file_monitor_remove (node->file, model);
 		path = get_node_path (model, node);
 
+		/* Report row_deleted before actually deleting */
+		gtk_tree_model_row_deleted (GTK_TREE_MODEL (model), path);
+		gtk_tree_path_free (path);
+		
 		if (node->prev) {
 			node->prev->next = node->next;
 		}
@@ -1624,12 +1639,9 @@ fm_tree_model_remove_root_uri (FMTreeModel *model, const char *uri)
 		
 		/* destroy the root identifier */
 		root = node->root;
-		g_signal_handler_disconnect (node->file, root->changed_handler_id);
 		destroy_node_without_reporting (model, node);
 		g_hash_table_destroy (root->file_to_node_map);
 		g_free (root);
-		gtk_tree_model_row_deleted (GTK_TREE_MODEL (model), path);
-		gtk_tree_path_free (path);
 	}
 }
 
@@ -1741,8 +1753,8 @@ fm_tree_model_iter_get_file (FMTreeModel *model, GtkTreeIter *iter)
 {
 	TreeNode *node;
 
-	g_return_val_if_fail (FM_IS_TREE_MODEL (model), 0);
-	g_return_val_if_fail (iter_is_valid (FM_TREE_MODEL (model), iter), 0);
+	g_return_val_if_fail (FM_IS_TREE_MODEL (model), NULL);
+	g_return_val_if_fail (iter_is_valid (FM_TREE_MODEL (model), iter), NULL);
 
 	node = iter->user_data;
 	return node == NULL ? NULL : nautilus_file_ref (node->file);
@@ -1807,8 +1819,6 @@ fm_tree_model_finalize (GObject *object)
 	for (root_node = model->details->root_node; root_node != NULL; root_node = next_root) {
 		next_root = root_node->next;
 		root = root_node->root;
-		g_signal_handler_disconnect (root_node->file, root->changed_handler_id);
-		nautilus_file_monitor_remove (root_node->file, model);
 		destroy_node_without_reporting (model, root_node);
 		g_hash_table_destroy (root->file_to_node_map);
 		g_free (root);
