@@ -31,6 +31,7 @@
 #include "nautilus-actions.h"
 #include "nautilus-application.h"
 #include "nautilus-location-bar.h"
+#include "nautilus-search-bar.h"
 #include "nautilus-pathbar.h"
 #include "nautilus-main.h"
 #include "nautilus-window-private.h"
@@ -46,19 +47,22 @@
 #include <gtk/gtkmain.h>
 #include <gtk/gtksignal.h>
 #include <gdk/gdkx.h>
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 #include <libgnomeui/gnome-dialog-util.h>
 #include <libgnomeui/gnome-icon-theme.h>
 #include <libgnomevfs/gnome-vfs-async-ops.h>
 #include <libgnomevfs/gnome-vfs-uri.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
+#include <libnautilus-extension/nautilus-location-widget-provider.h>
 #include <libnautilus-private/nautilus-file-attributes.h>
 #include <libnautilus-private/nautilus-file-utilities.h>
 #include <libnautilus-private/nautilus-file.h>
 #include <libnautilus-private/nautilus-global-preferences.h>
 #include <libnautilus-private/nautilus-metadata.h>
 #include <libnautilus-private/nautilus-mime-actions.h>
+#include <libnautilus-private/nautilus-module.h>
 #include <libnautilus-private/nautilus-monitor.h>
+#include <libnautilus-private/nautilus-search-directory.h>
 #include <libnautilus-private/nautilus-view-factory.h>
 #include <libnautilus-private/nautilus-window-info.h>
 
@@ -104,6 +108,8 @@ static void location_has_really_changed               (NautilusWindow
 static void update_for_new_location                   (NautilusWindow             *window);
 static void zoom_parameters_changed_callback          (NautilusView               *view,
 						       NautilusWindow             *window);
+static void update_extra_location_widgets_visibility  (NautilusWindow             *window);
+static void remove_extra_location_widgets             (NautilusWindow             *window);
 
 void
 nautilus_window_report_selection_changed (NautilusWindowInfo *window)
@@ -290,6 +296,7 @@ viewed_file_changed_callback (NautilusFile *file,
                               NautilusWindow *window)
 {
         char *new_location;
+	gboolean is_in_trash, was_in_trash;
 
         g_assert (NAUTILUS_IS_FILE (file));
 	g_assert (NAUTILUS_IS_WINDOW (window));
@@ -299,8 +306,12 @@ viewed_file_changed_callback (NautilusFile *file,
                 window->details->viewed_file_seen = TRUE;
         }
 
-	/* Close window if the file it's viewing has been deleted. */
-	if (nautilus_file_is_gone (file)) {
+	was_in_trash = window->details->viewed_file_in_trash;
+
+	window->details->viewed_file_in_trash = is_in_trash = nautilus_file_is_in_trash (file);
+
+	/* Close window if the file it's viewing has been deleted or moved to trash. */
+	if (nautilus_file_is_gone (file) || (is_in_trash && !was_in_trash)) {
                 /* Don't close the window in the case where the
                  * file was never seen in the first place.
                  */
@@ -319,12 +330,32 @@ viewed_file_changed_callback (NautilusFile *file,
                          * all other nautilus_window_close callers?
                          */
                         end_location_change (window);
-                        
-                        /* FIXME bugzilla.gnome.org 45038: Is closing
-                         * the window really the right thing to do for
-                         * all cases?
-                         */
-                        nautilus_window_close (window);
+
+			if (NAUTILUS_IS_NAVIGATION_WINDOW (window)) {
+				/* auto-show existing parent URI. */
+				char *uri, *go_to_uri;
+				NautilusFile *parent_file;
+
+				parent_file = nautilus_file_get_parent (file);
+				uri = nautilus_file_get_uri (parent_file);
+
+				go_to_uri = nautilus_find_existing_uri_in_hierarchy (uri);
+				if (go_to_uri != NULL) {
+					/* the path bar URI will be set to go_to_uri immediately
+					 * in begin_location_change, but we don't want the
+					 * inexistant children to show up anymore */
+					nautilus_path_bar_clear_buttons (NAUTILUS_PATH_BAR (NAUTILUS_NAVIGATION_WINDOW (window)->path_bar));
+					nautilus_window_go_to (NAUTILUS_WINDOW (window), go_to_uri);
+					g_free (go_to_uri);
+				} else {
+					nautilus_window_go_home (NAUTILUS_WINDOW (window));
+				}
+
+				g_free (uri);
+				nautilus_file_unref (parent_file);
+			} else {
+				nautilus_window_close (window);
+			}
                 }
 	} else {
                 new_location = nautilus_file_get_uri (file);
@@ -458,6 +489,11 @@ nautilus_window_open_location_full (NautilusWindow *window,
 				} else {
 					NAUTILUS_SPATIAL_WINDOW (window)->affect_spatial_window_on_next_location_change = FALSE;
 				}
+			} else if ((flags & NAUTILUS_WINDOW_OPEN_FLAG_NEW_WINDOW) != 0) {
+				target_window = nautilus_application_create_navigation_window 
+					(window->application,
+					 NULL,
+					 gtk_window_get_screen (GTK_WINDOW (window)));
 			}
 		} else if (NAUTILUS_IS_SPATIAL_WINDOW (window)) {
                         if (!NAUTILUS_SPATIAL_WINDOW (window)->affect_spatial_window_on_next_location_change) {
@@ -473,7 +509,12 @@ nautilus_window_open_location_full (NautilusWindow *window,
                                 NAUTILUS_SPATIAL_WINDOW (window)->affect_spatial_window_on_next_location_change = FALSE;
                                 target_window = window;
                         }
-                } else {
+		} else if (flags & NAUTILUS_WINDOW_OPEN_FLAG_NEW_WINDOW) {
+			target_window = nautilus_application_create_navigation_window 
+				(window->application,
+				 NULL,
+				 gtk_window_get_screen (GTK_WINDOW (window)));
+		} else {
                         target_window = window;
                 }       
                 break;
@@ -597,7 +638,6 @@ report_current_content_view_failure_to_user (NautilusWindow *window,
 	message = nautilus_window_get_view_startup_error_label (window);
 	eel_show_error_dialog (message,
 			       _("You can choose another view or go to a different location."),
-			       _("View Failed"),
 			       GTK_WINDOW (window));
 	g_free (message);
 }
@@ -611,7 +651,6 @@ report_nascent_content_view_failure_to_user (NautilusWindow *window,
 	message = nautilus_window_get_view_error_label (window);
 	eel_show_error_dialog (message,
 			       _("The location cannot be displayed with this viewer."),
-			       _("View Failed"),
 			       GTK_WINDOW (window));
 	g_free (message);
 }
@@ -638,31 +677,6 @@ nautilus_window_content_view_matches_iid (NautilusWindow *window,
 }
 
 
-static gboolean
-another_navigation_window_already_showing (NautilusWindow *the_window)
-{
-	GList *list, *item;
-	
-	list = nautilus_application_get_window_list ();
-	for (item = list; item != NULL; item = item->next) {
-		if (item->data != the_window &&
-		    NAUTILUS_IS_NAVIGATION_WINDOW (item->data)) {
-			return TRUE;
-		}
-	}
-	
-	return FALSE;
-}
-
-/* utility routine that returns true if there's one or fewer windows in the window list */
-static gboolean
-just_one_window (void)
-{
-	return !eel_g_list_more_than_one_item
-                (nautilus_application_get_window_list ());
-}
-
-
 /*
  * begin_location_change
  * 
@@ -751,7 +765,11 @@ begin_location_change (NautilusWindow *window,
         window->details->determine_view_file = nautilus_file_get (location);
 
 	g_assert (window->details->determine_view_file != NULL);
-	
+
+	/* if the currently viewed file is marked gone while loading the new location,
+	 * this ensures that the window isn't destroyed */
+        cancel_viewed_file_changed_callback (window);
+
 	nautilus_file_call_when_ready (window->details->determine_view_file,
 				       NAUTILUS_FILE_ATTRIBUTE_IS_DIRECTORY |
 				       NAUTILUS_FILE_ATTRIBUTE_MIME_TYPE |
@@ -768,6 +786,7 @@ setup_new_window (NautilusWindow *window, NautilusFile *file)
 	char *show_hidden_file_setting;
 	char *geometry_string;
 	char *scroll_string;
+	gboolean maximized;
 	
 	if (NAUTILUS_IS_SPATIAL_WINDOW (window) && !NAUTILUS_IS_DESKTOP_WINDOW (window)) {
 		/* load show hidden state */
@@ -786,6 +805,14 @@ setup_new_window (NautilusWindow *window, NautilusFile *file)
 		g_free (show_hidden_file_setting);
 		
 		/* load the saved window geometry */
+		maximized = nautilus_file_get_boolean_metadata
+			(file, NAUTILUS_METADATA_KEY_WINDOW_MAXIMIZED, FALSE);
+		if (maximized) {
+			gtk_window_maximize (GTK_WINDOW (window));
+		} else {
+			gtk_window_unmaximize (GTK_WINDOW (window));
+		}
+
 		geometry_string = nautilus_file_get_metadata 
 			(file, NAUTILUS_METADATA_KEY_WINDOW_GEOMETRY, NULL);
                 if (geometry_string != NULL) {
@@ -814,25 +841,6 @@ setup_new_window (NautilusWindow *window, NautilusFile *file)
 			window->details->pending_scroll_to = scroll_string;
 		}
         }
-	
-        if (NAUTILUS_IS_NAVIGATION_WINDOW (window)) {
-		geometry_string = eel_preferences_get
-				(NAUTILUS_PREFERENCES_NAVIGATION_WINDOW_SAVED_GEOMETRY);
-                if (geometry_string != NULL &&
-		    geometry_string[0] != 0) {
-			/* Ignore saved window position if a window with the same
-			 * location is already showing. That way the two windows
-			 * wont appear at the exact same location on the screen.
-			 */
-                        eel_gtk_window_set_initial_geometry_from_string 
-                                (GTK_WINDOW (window), 
-                                 geometry_string,
-                                 NAUTILUS_WINDOW_MIN_WIDTH, 
-                                 NAUTILUS_WINDOW_MIN_HEIGHT,
-				 another_navigation_window_already_showing (window));
-                }
-                g_free (geometry_string);
-	}
 }
 
 static void
@@ -843,6 +851,7 @@ got_file_info_for_view_selection_callback (NautilusFile *file,
 	char *view_id;
 	char *mimetype;
 	NautilusWindow *window;
+	NautilusFile *viewed_file;
 	char *location;
 	char *home_uri;
 
@@ -911,7 +920,9 @@ got_file_info_for_view_selection_callback (NautilusFile *file,
 			 * happens when a new window cannot display its initial URI. 
 			 */
 			/* if this is the only window, we don't want to quit, so we redirect it to home */
-			if (just_one_window ()) {
+			if (nautilus_application_get_n_windows () <= 1) {
+				g_assert (nautilus_application_get_n_windows () == 1);
+
 				/* Make sure we re-use this window */
 				if (NAUTILUS_IS_SPATIAL_WINDOW (window)) {
 					NAUTILUS_SPATIAL_WINDOW (window)->affect_spatial_window_on_next_location_change = TRUE;
@@ -939,6 +950,14 @@ got_file_info_for_view_selection_callback (NautilusFile *file,
 		} else {
 			/* Clean up state of already-showing window */
 			end_location_change (window);
+
+			/* We disconnected this, so we need to re-connect it */
+			viewed_file = nautilus_file_get (window->details->location);
+			nautilus_window_set_viewed_file (window, viewed_file);
+			nautilus_file_monitor_add (viewed_file, &window->details->viewed_file, 0);
+			g_signal_connect_object (viewed_file, "changed",
+						 G_CALLBACK (viewed_file_changed_callback), window, 0);
+			nautilus_file_unref (viewed_file);
 			
 			/* Leave the location bar showing the bad location that the user
 			 * typed (or maybe achieved by dragging or something). Many times
@@ -1125,12 +1144,36 @@ location_has_really_changed (NautilusWindow *window)
 	}
 }
 
+static void
+add_extension_extra_widgets (NautilusWindow *window, const char *uri)
+{
+	GList *providers, *l;
+	GtkWidget *widget;
+	
+	providers = nautilus_module_get_extensions_for_type (NAUTILUS_TYPE_LOCATION_WIDGET_PROVIDER);
+
+	for (l = providers; l != NULL; l = l->next) {
+		NautilusLocationWidgetProvider *provider;
+		
+		provider = NAUTILUS_LOCATION_WIDGET_PROVIDER (l->data);
+		widget = nautilus_location_widget_provider_get_widget (provider, uri, GTK_WIDGET (window));
+		if (widget != NULL) {
+			nautilus_window_add_extra_location_widget (window, widget);
+		}
+	}
+
+	nautilus_module_extension_list_free (providers);
+	
+}
+
 /* Handle the changes for the NautilusWindow itself. */
 static void
 update_for_new_location (NautilusWindow *window)
 {
         char *new_location;
         NautilusFile *file;
+	NautilusDirectory *directory;
+	gboolean location_really_changed;
         
         new_location = window->details->pending_location;
         window->details->pending_location = NULL;
@@ -1139,6 +1182,8 @@ update_for_new_location (NautilusWindow *window)
 
         update_history (window, window->details->location_change_type, new_location);
                 
+	location_really_changed = eel_strcmp (window->details->location, new_location) != 0;
+		
         /* Set the new location. */
         g_free (window->details->location);
         window->details->location = new_location;
@@ -1150,6 +1195,7 @@ update_for_new_location (NautilusWindow *window)
         file = nautilus_file_get (window->details->location);
         nautilus_window_set_viewed_file (window, file);
         window->details->viewed_file_seen = !nautilus_file_is_not_yet_confirmed (file);
+        window->details->viewed_file_in_trash = nautilus_file_is_in_trash (file);
         nautilus_file_monitor_add (file, &window->details->viewed_file, 0);
         g_signal_connect_object (file, "changed",
                                  G_CALLBACK (viewed_file_changed_callback), window, 0);
@@ -1168,6 +1214,22 @@ update_for_new_location (NautilusWindow *window)
 	/* Load menus from nautilus extensions for this location */
 	nautilus_window_load_extension_menus (window);
 
+	if (location_really_changed) {
+		remove_extra_location_widgets (window);
+		
+		directory = nautilus_directory_get (window->details->location);
+		if (NAUTILUS_IS_SEARCH_DIRECTORY (directory)) {
+			nautilus_window_set_search_mode (window, TRUE, NAUTILUS_SEARCH_DIRECTORY (directory));
+		} else {
+			nautilus_window_set_search_mode (window, FALSE, NULL);
+		}
+		nautilus_directory_unref (directory);
+
+		add_extension_extra_widgets (window, window->details->location);
+		
+		update_extra_location_widgets_visibility (window);
+	}
+
 #if !NEW_UI_COMPLETE
         if (NAUTILUS_IS_NAVIGATION_WINDOW (window)) {
                 /* Check if the back and forward buttons need enabling or disabling. */
@@ -1332,7 +1394,6 @@ display_view_selection_failure (NautilusWindow *window, NautilusFile *file,
 	char *error_message;
 	char *detail_message;
 	char *scheme_string;
-	char *dialog_title;
 	const char *host_name;
 	GtkDialog *dialog;
 	GnomeVFSURI *vfs_uri;
@@ -1348,9 +1409,7 @@ display_view_selection_failure (NautilusWindow *window, NautilusFile *file,
 	uri_for_display = eel_str_middle_truncate
 		(full_uri_for_display, MAX_URI_IN_DIALOG_LENGTH);
 	g_free (full_uri_for_display);
-	
-	dialog_title = NULL;
-        
+	        
 	switch (result_code) {
 	case GNOME_VFS_OK:
 		if (nautilus_file_is_directory (file)) {
@@ -1436,7 +1495,11 @@ display_view_selection_failure (NautilusWindow *window, NautilusFile *file,
 		detail_message = g_strdup
 			(_("Check that an SMB server is running in the local network."));
 		break;
-		
+
+	case GNOME_VFS_ERROR_CANCELLED:
+		g_free (uri_for_display);
+		return;
+
 	case GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE:
 	default:
 		error_message = g_strdup_printf (_("Nautilus cannot display \"%s\"."),
@@ -1444,13 +1507,8 @@ display_view_selection_failure (NautilusWindow *window, NautilusFile *file,
 		detail_message = g_strdup (_("Please select another viewer and try again."));
 	}
 	
-	if (dialog_title == NULL) {
-		dialog_title = g_strdup (_("Can't Display Location"));
-	}
-
-	dialog = eel_show_error_dialog (error_message, detail_message, dialog_title, NULL);
+	dialog = eel_show_error_dialog (error_message, detail_message, NULL);
 	
-	g_free (dialog_title);
 	g_free (uri_for_display);
 	g_free (error_message);
 	g_free (detail_message);
@@ -1532,10 +1590,7 @@ zoom_level_changed_callback (NautilusView *view,
 	action = gtk_action_group_get_action (window->details->main_action_group,
 					      NAUTILUS_ACTION_ZOOM_NORMAL);
 	gtk_action_set_visible (action, supports_zooming);
-	gtk_action_set_sensitive (action,
-				  TRUE);
-       
-	/* FIXME bugzilla.gnome.org 43442: Desensitize "Zoom Normal"? */
+	gtk_action_set_sensitive (action, supports_zooming);
 }
 
 static void
@@ -1700,3 +1755,44 @@ nautilus_window_reload (NautilusWindow *window)
 	g_free (location);
 	eel_g_list_free_deep (selection);
 }
+
+static void
+remove_all (GtkWidget *widget,
+	    gpointer data)
+{
+	GtkContainer *container;
+	container = GTK_CONTAINER (data);
+
+	gtk_container_remove (container, widget);
+}
+
+static void
+remove_extra_location_widgets (NautilusWindow *window)
+{
+	gtk_container_foreach (GTK_CONTAINER (window->details->extra_location_widgets),
+			       remove_all,
+			       window->details->extra_location_widgets);
+}
+
+void
+nautilus_window_add_extra_location_widget (NautilusWindow *window,
+					   GtkWidget *widget)
+{
+	gtk_box_pack_start (GTK_BOX (window->details->extra_location_widgets),
+			    widget, TRUE, TRUE, 0);
+}
+
+static void
+update_extra_location_widgets_visibility (NautilusWindow *window)
+{
+	GList *children;
+	
+	children = gtk_container_get_children (GTK_CONTAINER (window->details->extra_location_widgets));
+	
+	if (children != NULL) {
+		gtk_widget_show (window->details->extra_location_widgets);
+	} else {
+		gtk_widget_hide (window->details->extra_location_widgets);
+	}
+	g_list_free (children);
+}
